import uuid
from contextlib import contextmanager

import pytest

from dagster import PipelineDefinition, seven
from dagster.core.instance import DagsterInstance
from dagster.core.storage.pipeline_run import PipelineRun, PipelineRunStatus
from dagster.core.storage.runs import InMemoryRunStorage, SqliteRunStorage


def do_test_single_write_read(instance):
    run_id = 'some_run_id'
    pipeline_def = PipelineDefinition(name='some_pipeline', solid_defs=[])
    instance.create_empty_run(run_id=run_id, pipeline_name=pipeline_def.name)
    run = instance.get_run_by_id(run_id)
    assert run.run_id == run_id
    assert run.pipeline_name == 'some_pipeline'
    assert list(instance.all_runs()) == [run]
    instance.wipe()
    assert list(instance.all_runs()) == []


def build_run(
    run_id, pipeline_name, mode='default', tags=None, status=PipelineRunStatus.NOT_STARTED
):
    from dagster.core.definitions.pipeline import ExecutionSelector

    return PipelineRun(
        pipeline_name=pipeline_name,
        run_id=run_id,
        environment_dict=None,
        mode=mode,
        selector=ExecutionSelector(pipeline_name),
        reexecution_config=None,
        step_keys_to_execute=None,
        tags=tags,
        status=status,
    )


def test_filesystem_persist_one_run(tmpdir):
    do_test_single_write_read(DagsterInstance.local_temp(str(tmpdir)))


def test_in_memory_persist_one_run():
    do_test_single_write_read(DagsterInstance.ephemeral())


@contextmanager
def create_sqlite_run_storage():
    with seven.TemporaryDirectory() as tempdir:
        yield SqliteRunStorage.from_local(tempdir)


@contextmanager
def create_in_memory_storage():
    yield InMemoryRunStorage()


run_storage_test = pytest.mark.parametrize(
    'run_storage_factory_cm_fn', [create_sqlite_run_storage, create_in_memory_storage]
)


@run_storage_test
def test_basic_storage(run_storage_factory_cm_fn):
    with run_storage_factory_cm_fn() as storage:
        assert storage
        run_id = str(uuid.uuid4())
        added = storage.add_run(build_run(run_id=run_id, pipeline_name='some_pipeline'))
        assert added
        runs = storage.all_runs()
        assert len(runs) == 1
        run = runs[0]
        assert run.run_id == run_id
        assert run.pipeline_name == 'some_pipeline'
        assert storage.has_run(run_id)
        fetched_run = storage.get_run_by_id(run_id)
        assert fetched_run.run_id == run_id
        assert fetched_run.pipeline_name == 'some_pipeline'


@run_storage_test
def test_clear(run_storage_factory_cm_fn):
    with run_storage_factory_cm_fn() as storage:
        assert storage
        run_id = str(uuid.uuid4())
        storage.add_run(build_run(run_id=run_id, pipeline_name='some_pipeline'))
        assert len(storage.all_runs()) == 1
        storage.wipe()
        assert list(storage.all_runs()) == []


@run_storage_test
def test_fetch_by_pipeline(run_storage_factory_cm_fn):
    with run_storage_factory_cm_fn() as storage:
        assert storage
        one = str(uuid.uuid4())
        two = str(uuid.uuid4())
        storage.add_run(build_run(run_id=one, pipeline_name='some_pipeline'))
        storage.add_run(build_run(run_id=two, pipeline_name='some_other_pipeline'))
        assert len(storage.all_runs()) == 2
        some_runs = storage.get_runs_with_pipeline_name('some_pipeline')
        assert len(some_runs) == 1
        assert some_runs[0].run_id == one


@run_storage_test
def test_fetch_count_by_tag(run_storage_factory_cm_fn):
    with run_storage_factory_cm_fn() as storage:
        assert storage
        one = str(uuid.uuid4())
        two = str(uuid.uuid4())
        three = str(uuid.uuid4())
        storage.add_run(
            build_run(
                run_id=one,
                pipeline_name='some_pipeline',
                tags={'mytag': 'hello', 'mytag2': 'world'},
            )
        )
        storage.add_run(
            build_run(
                run_id=two,
                pipeline_name='some_pipeline',
                tags={'mytag': 'goodbye', 'mytag2': 'world'},
            )
        )
        storage.add_run(build_run(run_id=three, pipeline_name='some_pipeline'))
        assert len(storage.all_runs()) == 3

        run_count = storage.get_run_count_with_matching_tags(
            [('mytag', 'hello'), ('mytag2', 'world')]
        )
        assert run_count == 1

        run_count = storage.get_run_count_with_matching_tags([('mytag2', 'world')])
        assert run_count == 2

        run_count = storage.get_run_count_with_matching_tags([])
        assert run_count == 3


@run_storage_test
def test_fetch_by_tags(run_storage_factory_cm_fn):
    with run_storage_factory_cm_fn() as storage:
        assert storage
        one = str(uuid.uuid4())
        two = str(uuid.uuid4())
        three = str(uuid.uuid4())
        storage.add_run(
            build_run(
                run_id=one,
                pipeline_name='some_pipeline',
                tags={'mytag': 'hello', 'mytag2': 'world'},
            )
        )
        storage.add_run(
            build_run(
                run_id=two,
                pipeline_name='some_pipeline',
                tags={'mytag': 'goodbye', 'mytag2': 'world'},
            )
        )
        storage.add_run(build_run(run_id=three, pipeline_name='some_pipeline'))
        assert len(storage.all_runs()) == 3

        some_runs = storage.get_runs_with_matching_tags([('mytag', 'hello'), ('mytag2', 'world')])
        assert len(some_runs) == 1
        assert some_runs[0].run_id == one

        some_runs = storage.get_runs_with_matching_tags([('mytag2', 'world')])
        assert len(some_runs) == 2
        assert any(x.run_id == one for x in some_runs)
        assert any(x.run_id == two for x in some_runs)

        some_runs = storage.get_runs_with_matching_tags([])
        assert len(some_runs) == 3


@run_storage_test
def test_paginated_fetch(run_storage_factory_cm_fn):
    storage = InMemoryRunStorage()
    with run_storage_factory_cm_fn() as storage:
        assert storage
        one, two, three = [str(uuid.uuid4()), str(uuid.uuid4()), str(uuid.uuid4())]
        storage.add_run(
            build_run(run_id=one, pipeline_name='some_pipeline', tags={'mytag': 'hello'})
        )
        storage.add_run(
            build_run(run_id=two, pipeline_name='some_pipeline', tags={'mytag': 'hello'})
        )
        storage.add_run(
            build_run(run_id=three, pipeline_name='some_pipeline', tags={'mytag': 'hello'})
        )

        all_runs = storage.all_runs()
        assert len(all_runs) == 3
        sliced_runs = storage.all_runs(cursor=three, limit=1)
        assert len(sliced_runs) == 1
        assert sliced_runs[0].run_id == two

        all_runs = storage.get_runs_with_pipeline_name('some_pipeline')
        assert len(all_runs) == 3
        sliced_runs = storage.get_runs_with_pipeline_name('some_pipeline', cursor=three, limit=1)
        assert len(sliced_runs) == 1
        assert sliced_runs[0].run_id == two

        all_runs = storage.get_runs_with_matching_tags([('mytag', 'hello')])
        assert len(all_runs) == 3
        sliced_runs = storage.get_runs_with_matching_tags(
            [('mytag', 'hello')], cursor=three, limit=1
        )
        assert len(sliced_runs) == 1
        assert sliced_runs[0].run_id == two


@run_storage_test
def test_fetch_by_status(run_storage_factory_cm_fn):
    with run_storage_factory_cm_fn() as storage:
        assert storage
        one = str(uuid.uuid4())
        two = str(uuid.uuid4())
        three = str(uuid.uuid4())
        four = str(uuid.uuid4())
        storage.add_run(
            build_run(
                run_id=one, pipeline_name='some_pipeline', status=PipelineRunStatus.NOT_STARTED
            )
        )
        storage.add_run(
            build_run(run_id=two, pipeline_name='some_pipeline', status=PipelineRunStatus.STARTED)
        )
        storage.add_run(
            build_run(run_id=three, pipeline_name='some_pipeline', status=PipelineRunStatus.STARTED)
        )
        storage.add_run(
            build_run(run_id=four, pipeline_name='some_pipeline', status=PipelineRunStatus.FAILURE)
        )

        assert {
            run.run_id for run in storage.get_runs_with_status(PipelineRunStatus.NOT_STARTED)
        } == {one}

        assert {run.run_id for run in storage.get_runs_with_status(PipelineRunStatus.STARTED)} == {
            two,
            three,
        }

        assert {run.run_id for run in storage.get_runs_with_status(PipelineRunStatus.FAILURE)} == {
            four
        }

        assert {
            run.run_id for run in storage.get_runs_with_status(PipelineRunStatus.SUCCESS)
        } == set()


@run_storage_test
def test_fetch_by_status_cursored(run_storage_factory_cm_fn):
    with run_storage_factory_cm_fn() as storage:
        assert storage
        one = str(uuid.uuid4())
        two = str(uuid.uuid4())
        three = str(uuid.uuid4())
        four = str(uuid.uuid4())
        storage.add_run(
            build_run(run_id=one, pipeline_name='some_pipeline', status=PipelineRunStatus.STARTED)
        )
        storage.add_run(
            build_run(run_id=two, pipeline_name='some_pipeline', status=PipelineRunStatus.STARTED)
        )
        storage.add_run(
            build_run(
                run_id=three, pipeline_name='some_pipeline', status=PipelineRunStatus.NOT_STARTED
            )
        )
        storage.add_run(
            build_run(run_id=four, pipeline_name='some_pipeline', status=PipelineRunStatus.STARTED)
        )

        cursor_four_runs = storage.get_runs_with_status(PipelineRunStatus.STARTED, cursor=four)
        assert len(cursor_four_runs) == 2
        assert {run.run_id for run in cursor_four_runs} == {one, two}

        cursor_two_runs = storage.get_runs_with_status(PipelineRunStatus.STARTED, cursor=two)
        assert len(cursor_two_runs) == 1
        assert {run.run_id for run in cursor_two_runs} == {one}

        cursor_one_runs = storage.get_runs_with_status(PipelineRunStatus.STARTED, cursor=one)
        assert not cursor_one_runs

        cursor_four_limit_one = storage.get_runs_with_status(
            PipelineRunStatus.STARTED, cursor=four, limit=1
        )
        assert len(cursor_four_limit_one) == 1
        assert cursor_four_limit_one[0].run_id == two


@run_storage_test
def test_delete(run_storage_factory_cm_fn):
    with run_storage_factory_cm_fn() as storage:
        assert storage
        run_id = str(uuid.uuid4())
        storage.add_run(build_run(run_id=run_id, pipeline_name='some_pipeline'))
        assert len(storage.all_runs()) == 1
        storage.delete_run(run_id)
        assert list(storage.all_runs()) == []



from dagster import InputDefinition, Int, List, Optional, PipelineDefinition, lambda_solid
from dagster.core.types.runtime import ALL_RUNTIME_BUILTINS, resolve_to_runtime_type


def inner_type_key_set(runtime_type):
    return {t.key for t in runtime_type.inner_types}


def test_inner_types():
    assert resolve_to_runtime_type(Int).inner_types == []

    list_int_runtime = resolve_to_runtime_type(List[Int])
    assert inner_type_key_set(list_int_runtime) == set(['Int'])

    list_list_int_runtime = resolve_to_runtime_type(List[List[Int]])
    assert inner_type_key_set(list_list_int_runtime) == set(['Int', 'List.Int'])

    list_nullable_int_runtime = resolve_to_runtime_type(List[Optional[Int]])
    assert inner_type_key_set(list_nullable_int_runtime) == set(['Int', 'Optional.Int'])
    assert not list_nullable_int_runtime.is_scalar


def test_is_any():
    assert not resolve_to_runtime_type(Int).is_any
    assert resolve_to_runtime_type(Int).is_scalar


def test_display_name():

    int_runtime = resolve_to_runtime_type(Int)
    assert int_runtime.display_name == 'Int'
    list_int_runtime = resolve_to_runtime_type(List[Int])
    assert list_int_runtime.display_name == '[Int]'
    list_list_int_runtime = resolve_to_runtime_type(List[List[Int]])
    assert list_list_int_runtime.display_name == '[[Int]]'
    list_nullable_int_runtime = resolve_to_runtime_type(List[Optional[Int]])
    assert list_nullable_int_runtime.display_name == '[Int?]'


def test_builtins_available():
    pipeline = PipelineDefinition(name='test_builting_available', solid_defs=[])
    for builtin_type in ALL_RUNTIME_BUILTINS:
        assert pipeline.has_runtime_type(builtin_type.name)
        assert pipeline.runtime_type_named(builtin_type.name).is_builtin


def test_python_mapping():
    runtime = resolve_to_runtime_type(int)
    assert runtime.name == 'Int'
    runtime = resolve_to_runtime_type(str)
    assert runtime.name == 'String'
    runtime = resolve_to_runtime_type(bool)
    assert runtime.name == 'Bool'
    runtime = resolve_to_runtime_type(float)
    assert runtime.name == 'Float'

    @lambda_solid(input_defs=[InputDefinition('num', int)])
    def add_one(num):
        return num + 1

    assert add_one.input_defs[0].runtime_type.name == 'Int'

    runtime = resolve_to_runtime_type(float)
    runtime.type_check(1.0)
    res = runtime.type_check(1)
    assert not res.success

    runtime = resolve_to_runtime_type(bool)
    runtime.type_check(True)
    res = runtime.type_check(1)
    assert not res.success



import os
from collections import namedtuple

import yaml

from dagster import check
from dagster.core.serdes import ConfigurableClassData, whitelist_for_serdes

from .config import DAGSTER_CONFIG_YAML_FILENAME, dagster_instance_config


def _runs_directory(base):
    return os.path.join(base, 'history', '')


def _compute_logs_directory(base):
    return os.path.join(base, 'storage')


def _event_logs_directory(base):
    return os.path.join(base, 'history', 'runs', '')


def configurable_class_data_or_default(config_value, field_name, default):
    if config_value.get(field_name):
        return ConfigurableClassData(
            config_value[field_name]['module'],
            config_value[field_name]['class'],
            yaml.dump(config_value[field_name]['config'], default_flow_style=False),
        )
    return default


@whitelist_for_serdes
class InstanceRef(
    namedtuple(
        '_InstanceRef',
        'local_artifact_storage_data run_storage_data event_storage_data compute_logs_data '
        'run_launcher_data',
    )
):
    def __new__(
        self,
        local_artifact_storage_data,
        run_storage_data,
        event_storage_data,
        compute_logs_data,
        run_launcher_data,
    ):
        return super(self, InstanceRef).__new__(
            self,
            local_artifact_storage_data=check.inst_param(
                local_artifact_storage_data, 'local_artifact_storage_data', ConfigurableClassData
            ),
            run_storage_data=check.inst_param(
                run_storage_data, 'run_storage_data', ConfigurableClassData
            ),
            event_storage_data=check.inst_param(
                event_storage_data, 'event_storage_data', ConfigurableClassData
            ),
            compute_logs_data=check.inst_param(
                compute_logs_data, 'compute_logs_data', ConfigurableClassData
            ),
            run_launcher_data=check.opt_inst_param(
                run_launcher_data, 'run_launcher_data', ConfigurableClassData
            ),
        )

    @staticmethod
    def from_dir(base_dir, config_filename=DAGSTER_CONFIG_YAML_FILENAME, overrides=None):
        overrides = check.opt_dict_param(overrides, 'overrides')
        config_value = dagster_instance_config(
            base_dir, config_filename=config_filename, overrides=overrides
        )

        local_artifact_storage_data = configurable_class_data_or_default(
            config_value,
            'local_artifact_storage',
            ConfigurableClassData(
                'dagster.core.storage.root',
                'LocalArtifactStorage',
                yaml.dump({'base_dir': base_dir}, default_flow_style=False),
            ),
        )

        run_storage_data = configurable_class_data_or_default(
            config_value,
            'run_storage',
            ConfigurableClassData(
                'dagster.core.storage.runs',
                'SqliteRunStorage',
                yaml.dump({'base_dir': _runs_directory(base_dir)}, default_flow_style=False),
            ),
        )

        event_storage_data = configurable_class_data_or_default(
            config_value,
            'event_log_storage',
            ConfigurableClassData(
                'dagster.core.storage.event_log',
                'SqliteEventLogStorage',
                yaml.dump({'base_dir': _event_logs_directory(base_dir)}, default_flow_style=False),
            ),
        )

        compute_logs_data = configurable_class_data_or_default(
            config_value,
            'compute_logs',
            ConfigurableClassData(
                'dagster.core.storage.local_compute_log_manager',
                'LocalComputeLogManager',
                yaml.dump(
                    {'base_dir': _compute_logs_directory(base_dir)}, default_flow_style=False
                ),
            ),
        )

        run_launcher_data = configurable_class_data_or_default(config_value, 'run_launcher', None)

        return InstanceRef(
            local_artifact_storage_data=local_artifact_storage_data,
            run_storage_data=run_storage_data,
            event_storage_data=event_storage_data,
            compute_logs_data=compute_logs_data,
            run_launcher_data=run_launcher_data,
        )

    @property
    def local_artifact_storage(self):
        return self.local_artifact_storage_data.rehydrate()

    @property
    def run_storage(self):
        return self.run_storage_data.rehydrate()

    @property
    def event_storage(self):
        return self.event_storage_data.rehydrate()

    @property
    def compute_log_manager(self):
        return self.compute_logs_data.rehydrate()

    @property
    def run_launcher(self):
        return self.run_launcher_data.rehydrate() if self.run_launcher_data else None



import pytest

from dagster.utils.timing import TimerResult, format_duration, time_execution_scope


def test_format_duration():
    timing_pairs = [
        (0.0, "0.0ms"),
        (0.84123, "0.84ms"),
        (5.67822, "5.68ms"),
        (10.123, "10ms"),
        (533.12123, "533ms"),
        (4567.123123, "4.57s"),
        (320000.1232, "5m20s"),
        (3910056.123, "1h5m"),
        (9022312.1233, "2h30m"),
    ]

    for ms, time_str in timing_pairs:
        assert format_duration(ms) == time_str


def test_basic_usage():
    with time_execution_scope() as timer_result:
        pass

    assert timer_result
    assert isinstance(timer_result.millis, float)
    assert isinstance(timer_result.seconds, float)


def test_wrong_usage():
    with pytest.raises(Exception):
        timer_result = TimerResult()
        assert timer_result.millis

    with pytest.raises(Exception):
        timer_result = TimerResult()
        assert timer_result.seconds



import keyword
import re

from dagster import check
from dagster.core.errors import DagsterInvalidDefinitionError

DEFAULT_OUTPUT = 'result'

DISALLOWED_NAMES = set(
    [
        'context',
        'conf',
        'config',
        'meta',
        'arg_dict',
        'dict',
        'input_arg_dict',
        'output_arg_dict',
        'int',
        'str',
        'float',
        'bool',
        'input',
        'output',
        'type',
    ]
    + keyword.kwlist  # just disallow all python keywords
)

VALID_NAME_REGEX_STR = r'^[A-Za-z0-9_]+$'
VALID_NAME_REGEX = re.compile(VALID_NAME_REGEX_STR)


def has_valid_name_chars(name):
    return bool(VALID_NAME_REGEX.match(name))


def check_valid_name(name):
    check.str_param(name, 'name')
    if name in DISALLOWED_NAMES:
        raise DagsterInvalidDefinitionError('{name} is not allowed.'.format(name=name))

    if not has_valid_name_chars(name):
        raise DagsterInvalidDefinitionError(
            '{name} must be in regex {regex}'.format(name=name, regex=VALID_NAME_REGEX_STR)
        )
    return name


def _kv_str(key, value):
    return '{key}="{value}"'.format(key=key, value=repr(value))


def struct_to_string(name, **kwargs):
    # Sort the kwargs to ensure consistent representations across Python versions
    props_str = ', '.join([_kv_str(key, value) for key, value in sorted(kwargs.items())])
    return '{name}({props_str})'.format(name=name, props_str=props_str)



import logging

import coloredlogs

from dagster import seven
from dagster.core.definitions.logger import logger
from dagster.core.log_manager import coerce_valid_log_level
from dagster.core.types import Field
from dagster.utils.log import default_format_string


@logger(
    {
        'log_level': Field(str, is_optional=True, default_value='INFO'),
        'name': Field(str, is_optional=True, default_value='dagster'),
    },
    description='The default colored console logger.',
)
def colored_console_logger(init_context):
    level = coerce_valid_log_level(init_context.logger_config['log_level'])
    name = init_context.logger_config['name']

    klass = logging.getLoggerClass()
    logger_ = klass(name, level=level)
    coloredlogs.install(
        logger=logger_,
        level=level,
        fmt=default_format_string(),
        field_styles={'levelname': {'color': 'blue'}, 'asctime': {'color': 'green'}},
        level_styles={'debug': {}, 'error': {'color': 'red'}},
    )
    return logger_


@logger(
    {
        'log_level': Field(str, is_optional=True, default_value='INFO'),
        'name': Field(str, is_optional=True, default_value='dagster'),
    },
    description='A JSON-formatted console logger',
)
def json_console_logger(init_context):
    level = coerce_valid_log_level(init_context.logger_config['log_level'])
    name = init_context.logger_config['name']

    klass = logging.getLoggerClass()
    logger_ = klass(name, level=level)

    handler = coloredlogs.StandardErrorHandler()

    class JsonFormatter(logging.Formatter):
        def format(self, record):
            return seven.json.dumps(record.__dict__)

    handler.setFormatter(JsonFormatter())
    logger_.addHandler(handler)

    return logger_


def default_system_loggers():
    '''If users don't provide configuration for any loggers, we instantiate these loggers with the
    default config.

    Returns:
        List[Tuple[LoggerDefinition, dict]]: Default loggers and their associated configs.'''
    return [(colored_console_logger, {'name': 'dagster', 'log_level': 'DEBUG'})]


def default_loggers():
    return {'console': colored_console_logger}



from __future__ import print_function

from dagster import check
from dagster.utils.indenting_printer import IndentingPrinter

from .config import ConfigType
from .field import resolve_to_config_type


def print_type(config_type, print_fn=print, with_lines=True):
    check.inst_param(config_type, 'config_type', ConfigType)
    check.callable_param(print_fn, 'print_fn')

    if with_lines:
        printer = IndentingPrinter(printer=print_fn)
    else:
        printer = IndentingPrinter(printer=print_fn, indent_level=0)
    _do_print(config_type, printer, with_lines=with_lines)
    printer.line('')


def _do_print(config_type, printer, with_lines=True, shortcut_named=False):
    line_break_fn = printer.line if with_lines else lambda string: printer.append(string + ' ')

    if config_type.is_list:
        printer.append('[')
        _do_print(config_type.inner_type, printer)
        printer.append(']')
    elif config_type.is_nullable:
        _do_print(config_type.inner_type, printer)
        printer.append('?')
    elif config_type.has_fields:
        if config_type.name and shortcut_named:
            printer.append(config_type.name)
        else:
            line_break_fn('{')
            with printer.with_indent():
                for name, field in sorted(config_type.fields.items()):
                    if field.is_optional:
                        printer.append(name + '?: ')
                    else:
                        printer.append(name + ': ')
                    _do_print(
                        field.config_type, printer, with_lines=with_lines, shortcut_named=True
                    )
                    line_break_fn('')

            printer.append('}')
    elif config_type.name:
        printer.append(config_type.name)
    else:
        check.failed('not supported')


def print_type_to_string(dagster_type):
    config_type = resolve_to_config_type(dagster_type)
    return print_config_type_to_string(config_type)


def print_config_type_to_string(config_type, with_lines=True):
    prints = []

    def _push(text):
        prints.append(text)

    print_type(config_type, _push, with_lines=with_lines)

    if with_lines:
        return '\n'.join(prints)
    else:
        return ' '.join(prints)



from __future__ import print_function

import io
import os
import signal
import subprocess
import sys
import time
from contextlib import contextmanager

from dagster import check
from dagster.core.execution.context.system import SystemStepExecutionContext
from dagster.core.storage.compute_log_manager import ComputeIOType
from dagster.seven import IS_WINDOWS
from dagster.utils import ensure_file

WIN_PY36_COMPUTE_LOG_DISABLED_MSG = '''\u001b[33mWARNING: Compute log capture is disabled for the current environment. Set the environment variable `PYTHONLEGACYWINDOWSSTDIO` to enable.\n\u001b[0m'''


@contextmanager
def mirror_step_io(step_context):
    check.inst_param(step_context, 'step_context', SystemStepExecutionContext)
    manager = step_context.instance.compute_log_manager
    if not manager.enabled(step_context):
        yield
        return

    outpath = manager.get_local_path(
        step_context.run_id, step_context.step.key, ComputeIOType.STDOUT
    )
    errpath = manager.get_local_path(
        step_context.run_id, step_context.step.key, ComputeIOType.STDERR
    )

    manager.on_compute_start(step_context)
    with mirror_io(outpath, errpath):
        # compute function executed here
        yield
    manager.on_compute_finish(step_context)


def should_disable_io_stream_redirect():
    # See https://stackoverflow.com/a/52377087
    return (
        os.name == 'nt'
        and sys.version_info.major == 3
        and sys.version_info.minor >= 6
        and not os.environ.get('PYTHONLEGACYWINDOWSSTDIO')
    )


def warn_if_compute_logs_disabled():
    if should_disable_io_stream_redirect():
        print(WIN_PY36_COMPUTE_LOG_DISABLED_MSG)


@contextmanager
def mirror_io(outpath, errpath, buffering=1):
    with mirror_stream(outpath, ComputeIOType.STDOUT, buffering):
        with mirror_stream(errpath, ComputeIOType.STDERR, buffering):
            yield


@contextmanager
def mirror_stream(path, io_type, buffering=1):
    ensure_file(path)
    from_stream = sys.stderr if io_type == ComputeIOType.STDERR else sys.stdout
    with tailf(path, io_type):
        with open(path, 'a+', buffering=buffering) as to_stream:
            with redirect_stream(to_stream=to_stream, from_stream=from_stream):
                yield


@contextmanager
def redirect_stream(to_stream=os.devnull, from_stream=sys.stdout):
    # swap the file descriptors to capture system-level output in the process
    # From https://stackoverflow.com/questions/4675728/redirect-stdout-to-a-file-in-python/22434262#22434262
    from_fd = _fileno(from_stream)
    to_fd = _fileno(to_stream)

    if not from_fd or not to_fd or should_disable_io_stream_redirect():
        yield
        return

    with os.fdopen(os.dup(from_fd), 'wb') as copied:
        from_stream.flush()
        try:
            os.dup2(_fileno(to_stream), from_fd)
        except ValueError:
            with open(to_stream, 'wb') as to_file:
                os.dup2(to_file.fileno(), from_fd)
        try:
            yield from_stream
        finally:
            from_stream.flush()
            to_stream.flush()
            os.dup2(copied.fileno(), from_fd)


POLLING_INTERVAL = 0.1


@contextmanager
def tailf(path, io_type=ComputeIOType.STDOUT):
    if IS_WINDOWS:
        with execute_windows_tail(path, io_type):
            yield
    else:
        with execute_posix_tail(path, io_type):
            yield


@contextmanager
def execute_windows_tail(path, io_type):
    # Cannot use multiprocessing here because we already may be in a daemonized process
    # Instead, invoke a thin wrapper around tail_polling using the dagster cli
    cmd = '{} -m dagster utils tail {} --parent-pid {} --io-type {}'.format(
        sys.executable, path, os.getpid(), io_type
    ).split(' ')
    tail_process = subprocess.Popen(cmd)

    try:
        yield
    finally:
        if tail_process:
            time.sleep(2 * POLLING_INTERVAL)
            tail_process.terminate()


@contextmanager
def execute_posix_tail(path, io_type):
    cmd = 'tail -F -c +0 {}'.format(path).split(' ')

    # open a subprocess to tail the file and print to stdout
    stream = sys.stdout if io_type == ComputeIOType.STDOUT else sys.stderr
    stream = stream if _fileno(stream) else None
    tail_process = subprocess.Popen(cmd, stdout=stream)

    # fork a child watcher process to sleep/wait for the parent process (which yields to the
    # compute function) to either A) complete and clean-up OR B) segfault / die silently.  In
    # the case of B, the spawned tail process will not automatically get terminated, so we need
    # to make sure that we terminate it explicitly and then exit.
    watcher_pid = os.fork()

    def clean(*_args):
        try:
            if tail_process:
                tail_process.terminate()
        except OSError:
            pass
        try:
            if watcher_pid:
                os.kill(watcher_pid, signal.SIGTERM)
        except OSError:
            pass

    if watcher_pid == 0:
        # this is the child watcher process, sleep until orphaned, then kill the tail process
        # and exit
        while True:
            if os.getppid() == 1:  # orphaned process
                clean()
                os._exit(0)  # pylint: disable=W0212
            else:
                time.sleep(1)
    else:
        # this is the parent process, yield to the compute function and then terminate both the
        # tail and watcher processes.
        try:
            yield
        finally:
            clean()


def tail_polling(filepath, stream=sys.stdout, parent_pid=None):
    '''
    Tails a file and outputs the content to the specified stream via polling.
    The pid of the parent process (if provided) is checked to see if the tail process should be
    terminated, in case the parent is hard-killed / segfaults
    '''
    with open(filepath, 'r') as file:
        for block in iter(lambda: file.read(1024), None):
            if block:
                print(block, end='', file=stream)
            else:
                if parent_pid and current_process_is_orphaned(parent_pid):
                    sys.exit()
                time.sleep(POLLING_INTERVAL)


def current_process_is_orphaned(parent_pid):
    parent_pid = int(parent_pid)
    if sys.platform == 'win32':
        import psutil

        try:
            parent = psutil.Process(parent_pid)
            return parent.status() != psutil.STATUS_RUNNING
        except psutil.NoSuchProcess:
            return True

    else:
        return os.getppid() != parent_pid


def _fileno(stream):
    try:
        fd = getattr(stream, 'fileno', lambda: stream)()
    except io.UnsupportedOperation:
        # Test CLI runners will stub out stdout to a non-file stream, which will raise an
        # UnsupportedOperation if `fileno` is accessed.  We need to make sure we do not error out,
        # or tests will fail
        return None

    if isinstance(fd, int):
        return fd

    return None



from dagster import check
from dagster.utils import ensure_single_item

from .config import ConfigType


def apply_default_values(config_type, config_value):
    check.inst_param(config_type, 'config_type', ConfigType)

    if config_type.is_scalar:
        return config_value
    elif config_type.is_enum:
        return config_type.to_python_value(config_value)
    elif config_type.is_selector:
        return apply_default_values_to_selector(config_type, config_value)
    elif config_type.is_composite:
        return apply_defaults_to_composite_type(config_type, config_value)
    elif config_type.is_list:
        return apply_defaults_to_list_type(config_type, config_value)
    elif config_type.is_nullable:
        if config_value is None:
            return None
        return apply_default_values(config_type.inner_type, config_value)
    elif config_type.is_any:
        return config_value
    else:
        check.failed('Unsupported type {name}'.format(name=config_type.name))


def apply_default_values_to_selector(selector_type, config_value):
    check.param_invariant(
        selector_type.is_selector, 'selector_type', 'Non-selector not caught in validation'
    )

    if config_value:
        check.invariant(config_value and len(config_value) == 1)
        field_name, incoming_field_value = ensure_single_item(config_value)
    else:
        field_name, field_def = ensure_single_item(selector_type.fields)
        incoming_field_value = field_def.default_value if field_def.default_provided else None

    parent_field = selector_type.fields[field_name]
    field_value = apply_default_values(parent_field.config_type, incoming_field_value)
    return {field_name: field_value}


def apply_defaults_to_composite_type(composite_type, config_value):
    check.param_invariant(composite_type.is_composite, 'composite_type')
    config_value = check.opt_dict_param(config_value, 'config_value', key_type=str)

    fields = composite_type.fields
    incoming_fields = set(config_value.keys())

    processed_fields = {}

    for expected_field, field_def in fields.items():
        if expected_field in incoming_fields:
            processed_fields[expected_field] = apply_default_values(
                field_def.config_type, config_value[expected_field]
            )

        elif field_def.default_provided:
            processed_fields[expected_field] = field_def.default_value

        elif not field_def.is_optional:
            check.failed('Missing non-optional composite member not caught in validation')

    # For permissive composite fields, we skip applying defaults because these fields are unknown
    # to us
    if composite_type.is_permissive_composite:
        defined_fields = set(fields.keys())
        extra_fields = incoming_fields - defined_fields
        for extra_field in extra_fields:
            processed_fields[extra_field] = config_value[extra_field]

    return processed_fields


def apply_defaults_to_list_type(list_type, config_value):
    check.param_invariant(list_type.is_list, 'list_type')

    if not config_value:
        return []

    if not list_type.inner_type.is_nullable:
        if any((cv is None for cv in config_value)):
            check.failed('Null list member not caught in validation')

    return [apply_default_values(list_type.inner_type, item) for item in config_value]



from dagster import check
from dagster.core.types.field_utils import check_user_facing_opt_field_param

from .config import resolve_config_field


class LoggerDefinition(object):
    '''Core class for defining loggers.
    
    Loggers are pipeline-scoped logging handlers, which will be automatically invoked whenever
    solids in a pipeline log messages.

    Args:
        logger_fn (Callable[[InitLoggerContext], logging.Logger]): User-provided function to
            instantiate the logger. This logger will be automatically invoked whenever the methods
            on ``context.log`` are called from within solid compute logic.
        config_field (Optional[Field]): The type for the configuration data for this logger, if
            any. The config data will be available on the ``init_context`` passed to ``logger_fn``
            as ``init_context.logger_config``.
        description (Optional[str]): A human-readable description of this logger.
    '''

    def __init__(self, logger_fn, config_field=None, description=None):
        self._logger_fn = check.callable_param(logger_fn, 'logger_fn')
        self._config_field = check_user_facing_opt_field_param(
            config_field, 'config_field', 'of a LoggerDefinition or @logger'
        )
        self._description = check.opt_str_param(description, 'description')

    @property
    def logger_fn(self):
        return self._logger_fn

    @property
    def config_field(self):
        return self._config_field

    @property
    def description(self):
        return self._description


def logger(config=None, config_field=None, description=None):
    '''Define a logger.
    
    The decorated function should accept an :py:class:`InitLoggerContext` and return an instance of
    :py:class:`python:logging.Logger`. This function will become the ``logger_fn`` of an underlying
    :py:class:`LoggerDefinition`.

    Args:
        config (Optional[Dict[str, Field]]): The schema for the configuration data made available on
            the ``init_context`` passed to the decorated function (as
            ``init_context.logger_config``).
        config_field (Optional[Field]): Used in the rare case of a top level config type other than
            a dictionary.

            Only one of config or config_field can be provided.
        description (Optional[str]): A human-readable description of the logger.
    '''
    # This case is for when decorator is used bare, without arguments.
    # E.g. @logger versus @logger()
    if callable(config):
        return LoggerDefinition(logger_fn=config)

    config_field = resolve_config_field(config_field, config, '@logger')

    def _wrap(logger_fn):
        return LoggerDefinition(logger_fn, config_field, description)

    return _wrap



from collections import namedtuple

from dagster import check
from dagster.core.types.config import ConfigType

from .environment_configs import (
    EnvironmentClassCreationData,
    construct_config_type_dictionary,
    define_environment_cls,
)
from .pipeline import PipelineDefinition


class EnvironmentSchema(
    namedtuple(
        '_EnvironmentSchema', 'environment_type config_type_dict_by_name config_type_dict_by_key'
    )
):
    def __new__(cls, environment_type, config_type_dict_by_name, config_type_dict_by_key):
        return super(EnvironmentSchema, cls).__new__(
            cls,
            environment_type=check.inst_param(environment_type, 'environment_type', ConfigType),
            config_type_dict_by_name=check.dict_param(
                config_type_dict_by_name,
                'config_type_dict_by_name',
                key_type=str,
                value_type=ConfigType,
            ),
            config_type_dict_by_key=check.dict_param(
                config_type_dict_by_key,
                'config_type_dict_by_key',
                key_type=str,
                value_type=ConfigType,
            ),
        )

    def has_config_type(self, name):
        check.str_param(name, 'name')
        return name in self.config_type_dict_by_name

    def config_type_named(self, name):
        check.str_param(name, 'name')
        return self.config_type_dict_by_name[name]

    def config_type_keyed(self, key):
        check.str_param(key, 'key')
        return self.config_type_dict_by_key[key]

    def all_config_types(self):
        return self.config_type_dict_by_key.values()


def create_environment_schema(pipeline_def, mode=None):
    check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition)
    mode = check.opt_str_param(mode, 'mode', default=pipeline_def.get_default_mode_name())
    mode_definition = pipeline_def.get_mode_definition(mode)

    environment_cls = define_environment_cls(
        EnvironmentClassCreationData(
            pipeline_name=pipeline_def.name,
            solids=pipeline_def.solids,
            dependency_structure=pipeline_def.dependency_structure,
            mode_definition=mode_definition,
            logger_defs=mode_definition.loggers,
        )
    )

    environment_type = environment_cls.inst()

    config_type_dict_by_name, config_type_dict_by_key = construct_config_type_dictionary(
        pipeline_def.all_solid_defs, environment_type
    )

    return EnvironmentSchema(
        environment_type=environment_type,
        config_type_dict_by_name=config_type_dict_by_name,
        config_type_dict_by_key=config_type_dict_by_key,
    )


def create_environment_type(pipeline_def, mode=None):
    return create_environment_schema(pipeline_def, mode).environment_type



from collections import namedtuple

from dagster import check
from dagster.core.serdes import whitelist_for_serdes
from dagster.core.types.config import ConfigType, ConfigTypeKind
from dagster.core.types.field import Field


@whitelist_for_serdes
class NonGenericTypeRefMeta(namedtuple('_NonGenericTypeRefMeta', 'key')):
    def __new__(cls, key):
        return super(NonGenericTypeRefMeta, cls).__new__(cls, check.str_param(key, 'key'))


@whitelist_for_serdes
class ConfigTypeMeta(
    namedtuple(
        '_ConfigTypeMeta',
        'kind key name description is_builtin is_system_config '
        'type_param_refs '  # only valid for closed generics (Set, Tuple, List, Optional)
        'enum_values '  # only valid for enums
        'fields',  # only valid for dicts and selectors
    )
):
    def __new__(
        cls,
        kind,
        key,
        name,
        type_param_refs,
        enum_values,
        fields,
        description,
        is_builtin,
        is_system_config,
    ):
        return super(ConfigTypeMeta, cls).__new__(
            cls,
            kind=check.inst_param(kind, 'kind', ConfigTypeKind),
            key=check.str_param(key, 'key'),
            name=check.opt_str_param(name, 'name'),
            type_param_refs=None
            if type_param_refs is None
            else check.list_param(type_param_refs, 'type_param_refs', of_type=TypeRef),
            enum_values=None
            if enum_values is None
            else check.list_param(enum_values, 'enum_values', of_type=ConfigEnumValueMeta),
            fields=None
            if fields is None
            else check.list_param(fields, 'field', of_type=ConfigFieldMeta),
            description=check.opt_str_param(description, 'description'),
            is_builtin=check.bool_param(is_builtin, 'is_builtin'),
            is_system_config=check.bool_param(is_system_config, 'is_system_config'),
        )

    @property
    def inner_type_refs(self):
        '''
        This recurses through the type references with non-generic types as leaves.
        '''

        def _doit():
            next_level_refs = _get_next_level_refs(self)
            if next_level_refs:
                for next_level in next_level_refs:
                    for inner_ref in _recurse_through_generics(next_level):
                        yield inner_ref

        # there might be duplicate keys (esp for scalars)
        refs_by_key = {}

        for ref in _doit():
            if ref.key not in refs_by_key:
                refs_by_key[ref.key] = ref

        return list(refs_by_key.values())


# This function is used by the recursive descent
# through all the inner types. This does *not*
# recursively descend through the type parameters
# of generic types. It just gets the next level of
# types. Either the direct type parameters of a
# generic type. Or the type refs of all the fields
# if it is a type with fields.
def _get_next_level_refs(ref):
    # if a generic type, get type params
    # if a type with fields, get refs of the fields
    if ConfigTypeKind.is_closed_generic(ref.kind):
        return ref.type_param_refs
    elif (
        ConfigTypeKind.has_fields(ref.kind) and ref.fields
    ):  # still check fields because permissive
        return [field_meta.type_ref for field_meta in ref.fields]


def _recurse_through_generics(ref):
    yield ref
    if isinstance(ref, ConfigTypeMeta) and ConfigTypeKind.is_closed_generic(ref.kind):
        for type_param_ref in ref.type_param_refs:
            for inner_ref in _recurse_through_generics(type_param_ref):
                yield inner_ref


# A type reference in these serializable data structures are one of two things
# 1) A closed generic type (e.g. List[Int] of Optional[Set[str]])
# 2) Or a reference to a non-generic type, such as Dict, Selector, or a Scalar.
# Upon deserialization and when hydrated back to the graphql query, it will
# be the responsibility of that module to maintain a dictionary of the
# non-generic types and then do lookups into the dictionary in order to
# to explode the entire type hierarchy requested by the client

TypeRef = (ConfigTypeMeta, NonGenericTypeRefMeta)


@whitelist_for_serdes
class ConfigEnumValueMeta(namedtuple('_ConfigEnumValueMeta', 'value description')):
    def __new__(cls, value, description):
        return super(ConfigEnumValueMeta, cls).__new__(
            cls,
            value=check.str_param(value, 'value'),
            description=check.opt_str_param(description, 'description'),
        )


@whitelist_for_serdes
class ConfigFieldMeta(
    namedtuple(
        '_ConfigFieldMeta',
        'name type_ref is_optional default_provided default_value_as_str description',
    )
):
    def __new__(
        cls, name, type_ref, is_optional, default_provided, default_value_as_str, description
    ):
        return super(ConfigFieldMeta, cls).__new__(
            cls,
            name=check.opt_str_param(name, 'name'),
            type_ref=check.inst_param(type_ref, 'type_ref', TypeRef),
            is_optional=check.bool_param(is_optional, 'is_optional'),
            default_provided=check.bool_param(default_provided, 'default_provided'),
            default_value_as_str=check.opt_str_param(default_value_as_str, 'default_value_as_str'),
            description=check.opt_str_param(description, 'description'),
        )


def meta_from_field(name, field):
    check.str_param(name, 'name')
    check.inst_param(field, 'field', Field)
    return ConfigFieldMeta(
        name=name,
        type_ref=type_ref_of(field.config_type),
        is_optional=field.is_optional,
        default_provided=field.default_provided,
        default_value_as_str=field.default_value_as_str if field.default_provided else None,
        description=field.description,
    )


def type_ref_of(config_type):
    check.inst_param(config_type, 'config_type', ConfigType)
    if ConfigTypeKind.is_closed_generic(config_type.kind):
        return meta_from_config_type(config_type)
    else:
        return NonGenericTypeRefMeta(key=config_type.key)


def type_refs_of(type_list):
    return list(map(type_ref_of, type_list)) if type_list is not None else None


def meta_from_config_type(config_type):
    check.inst_param(config_type, 'config_type', ConfigType)
    return ConfigTypeMeta(
        key=config_type.key,
        name=config_type.name,
        kind=config_type.kind,
        description=config_type.description,
        is_builtin=config_type.is_builtin,
        is_system_config=config_type.is_system_config,
        type_param_refs=type_refs_of(config_type.type_params),
        enum_values=[
            ConfigEnumValueMeta(ev.config_value, ev.description) for ev in config_type.enum_values
        ]
        if config_type.is_enum
        else None,
        fields=[meta_from_field(name, field) for name, field in config_type.fields.items()]
        if config_type.is_selector or config_type.is_composite
        else None,
    )



from dagster import (
    DependencyDefinition,
    ExecutionTargetHandle,
    InputDefinition,
    PipelineDefinition,
    execute_pipeline,
    lambda_solid,
)
from dagster.core.instance import DagsterInstance


def test_diamond_simple_execution():
    result = execute_pipeline(define_diamond_pipeline())
    assert result.success
    assert result.result_for_solid('adder').output_value() == 11


def compute_event(result, solid_name):
    return result.result_for_solid(solid_name).compute_step_events[0]


def test_diamond_multi_execution():
    pipeline = ExecutionTargetHandle.for_pipeline_python_file(
        __file__, 'define_diamond_pipeline'
    ).build_pipeline_definition()
    result = execute_pipeline(
        pipeline,
        environment_dict={'storage': {'filesystem': {}}, 'execution': {'multiprocess': {}}},
        instance=DagsterInstance.local_temp(),
    )
    assert result.success

    assert result.result_for_solid('adder').output_value() == 11

    # https://github.com/dagster-io/dagster/issues/1875
    # pids_by_solid = {}
    # for solid in pipeline.solids:
    #     pids_by_solid[solid.name] = compute_event(result, solid.name).logging_tags['pid']

    # # guarantee that all solids ran in their own process
    # assert len(set(pids_by_solid.values())) == len(pipeline.solids)


def define_diamond_pipeline():
    @lambda_solid
    def return_two():
        return 2

    @lambda_solid(input_defs=[InputDefinition('num')])
    def add_three(num):
        return num + 3

    @lambda_solid(input_defs=[InputDefinition('num')])
    def mult_three(num):
        return num * 3

    @lambda_solid(input_defs=[InputDefinition('left'), InputDefinition('right')])
    def adder(left, right):
        return left + right

    return PipelineDefinition(
        name='diamond_execution',
        solid_defs=[return_two, add_three, mult_three, adder],
        dependencies={
            'add_three': {'num': DependencyDefinition('return_two')},
            'mult_three': {'num': DependencyDefinition('return_two')},
            'adder': {
                'left': DependencyDefinition('add_three'),
                'right': DependencyDefinition('mult_three'),
            },
        },
    )


def define_error_pipeline():
    @lambda_solid
    def throw_error():
        raise Exception('bad programmer')

    return PipelineDefinition(name='error_pipeline', solid_defs=[throw_error])


def test_error_pipeline():
    pipeline = define_error_pipeline()
    result = execute_pipeline(pipeline, raise_on_error=False)
    assert not result.success


def test_error_pipeline_multiprocess():
    result = execute_pipeline(
        ExecutionTargetHandle.for_pipeline_fn(define_error_pipeline).build_pipeline_definition(),
        environment_dict={'storage': {'filesystem': {}}, 'execution': {'multiprocess': {}}},
        instance=DagsterInstance.local_temp(),
    )
    assert not result.success


def test_mem_storage_error_pipeline_multiprocess():
    result = execute_pipeline(
        ExecutionTargetHandle.for_pipeline_fn(define_diamond_pipeline).build_pipeline_definition(),
        environment_dict={'execution': {'multiprocess': {}}},
        instance=DagsterInstance.local_temp(),
        raise_on_error=False,
    )
    assert not result.success
    assert len(result.event_list) == 1
    assert result.event_list[0].is_failure


def test_invalid_instance():
    result = execute_pipeline(
        ExecutionTargetHandle.for_pipeline_fn(define_diamond_pipeline).build_pipeline_definition(),
        environment_dict={'storage': {'filesystem': {}}, 'execution': {'multiprocess': {}}},
        instance=DagsterInstance.ephemeral(),
        raise_on_error=False,
    )
    assert not result.success
    assert len(result.event_list) == 1
    assert result.event_list[0].is_failure



from dagster import check
from dagster.core.definitions.executor import ExecutorDefinition, default_executors
from dagster.loggers import default_loggers

from .logger import LoggerDefinition
from .resource import ResourceDefinition

DEFAULT_MODE_NAME = 'default'


class ModeDefinition(object):
    '''Define a mode in which a pipeline can operate.

    A mode provides pipelines with a set of resource implementations, loggers, system storages,
    and executors.

    Args:
        name (Optional[str]): The name of the mode. Must be unique within the
            :py:class:`PipelineDefinition` to which the mode is attached. (default: "default").
        resource_defs (Optional[Dict[str, ResourceDefinition]]): A dictionary of string resource
            keys to their implementations. Individual solids may require resources to be present by
            these keys.
        logger_defs (Optional[Dict[str, LoggerDefinition]]): A ditionary of string logger
            identifiers to their implementations.
        system_storage_defs (Optional[List[SystemStorageDefinition]]): The set of system storage
            options available when executing in this mode. By default, this will be the 'in_memory'
            and 'filesystem' system storages.
        executor_defs (Optional[List[ExecutorDefinition]]): The set of executors available when
            executing in this mode. By default, this will be the 'in_process' and 'multiprocess'
            executors.
        description (Optional[str]): A human-readable description of the mode.
    '''

    def __init__(
        self,
        name=None,
        resource_defs=None,
        logger_defs=None,
        system_storage_defs=None,
        executor_defs=None,
        description=None,
    ):
        from dagster.core.storage.system_storage import default_system_storage_defs

        from .system_storage import SystemStorageDefinition

        self.name = check.opt_str_param(name, 'name', DEFAULT_MODE_NAME)
        self.resource_defs = check.opt_dict_param(
            resource_defs, 'resource_defs', key_type=str, value_type=ResourceDefinition
        )
        self.loggers = (
            check.opt_dict_param(
                logger_defs, 'logger_defs', key_type=str, value_type=LoggerDefinition
            )
            or default_loggers()
        )
        self.system_storage_defs = check.list_param(
            system_storage_defs if system_storage_defs else default_system_storage_defs,
            'system_storage_defs',
            of_type=SystemStorageDefinition,
        )
        self.executor_defs = check.list_param(
            executor_defs if executor_defs else default_executors,
            'executor_defs',
            of_type=ExecutorDefinition,
        )
        self.description = check.opt_str_param(description, 'description')

    def get_system_storage_def(self, name):
        check.str_param(name, 'name')
        for system_storage_def in self.system_storage_defs:
            if system_storage_def.name == name:
                return system_storage_def

        check.failed('{} storage definition not found'.format(name))

    @staticmethod
    def from_resources(resources, name=None):
        check.dict_param(resources, 'resources', key_type=str)

        return ModeDefinition(
            name=name,
            resource_defs={
                resource_name: ResourceDefinition.hardcoded_resource(resource)
                for resource_name, resource in resources.items()
            },
        )



import re

import pytest

from dagster import (
    DagsterInvalidDefinitionError,
    DependencyDefinition,
    Dict,
    Field,
    InputDefinition,
    Int,
    OutputDefinition,
    PermissiveDict,
    PipelineDefinition,
    ResourceDefinition,
    SolidDefinition,
    String,
    solid,
)
from dagster.core.definitions import create_environment_schema
from dagster.core.types import NamedSelector, Selector
from dagster.core.types.config import ConfigTypeKind
from dagster.core.types.field_utils import NamedDict, _ConfigHasFields
from dagster.core.utility_solids import define_stub_solid


def solid_a_b_list():
    return [
        SolidDefinition(
            name='A',
            input_defs=[],
            output_defs=[OutputDefinition()],
            compute_fn=lambda _context, _inputs: None,
        ),
        SolidDefinition(
            name='B',
            input_defs=[InputDefinition('b_input')],
            output_defs=[],
            compute_fn=lambda _context, _inputs: None,
        ),
    ]


def test_create_pipeline_with_bad_solids_list():
    with pytest.raises(
        DagsterInvalidDefinitionError,
        match='"solids" arg to pipeline "a_pipeline" is not a list. Got',
    ):
        PipelineDefinition(
            name='a_pipeline', solid_defs=define_stub_solid('stub', [{'a key': 'a value'}])
        )


def test_circular_dep():
    with pytest.raises(DagsterInvalidDefinitionError, match='circular reference'):
        PipelineDefinition(
            solid_defs=solid_a_b_list(),
            dependencies={'A': {}, 'B': {'b_input': DependencyDefinition('B')}},
        )


def test_from_solid_not_there():
    with pytest.raises(
        DagsterInvalidDefinitionError, match='solid "NOTTHERE" in dependency dictionary not found'
    ):
        PipelineDefinition(
            solid_defs=solid_a_b_list(),
            dependencies={
                'A': {},
                'B': {'b_input': DependencyDefinition('A')},
                'NOTTHERE': {'b_input': DependencyDefinition('A')},
            },
        )


def test_from_non_existant_input():
    with pytest.raises(
        DagsterInvalidDefinitionError, match='solid "B" does not have input "not_an_input"'
    ):
        PipelineDefinition(
            solid_defs=solid_a_b_list(),
            dependencies={'B': {'not_an_input': DependencyDefinition('A')}},
        )


def test_to_solid_not_there():
    with pytest.raises(
        DagsterInvalidDefinitionError, match='solid "NOTTHERE" not found in solid list'
    ):
        PipelineDefinition(
            solid_defs=solid_a_b_list(),
            dependencies={'A': {}, 'B': {'b_input': DependencyDefinition('NOTTHERE')}},
        )


def test_to_solid_output_not_there():
    with pytest.raises(
        DagsterInvalidDefinitionError, match='solid "A" does not have output "NOTTHERE"'
    ):
        PipelineDefinition(
            solid_defs=solid_a_b_list(),
            dependencies={'B': {'b_input': DependencyDefinition('A', output='NOTTHERE')}},
        )


def test_invalid_item_in_solid_list():
    with pytest.raises(
        DagsterInvalidDefinitionError, match="Invalid item in solid list: 'not_a_solid'"
    ):
        PipelineDefinition(solid_defs=['not_a_solid'])


def test_one_layer_off_dependencies():
    with pytest.raises(
        DagsterInvalidDefinitionError,
        match="Received a IDependencyDefinition one layer too high under key B",
    ):
        PipelineDefinition(
            solid_defs=solid_a_b_list(), dependencies={'B': DependencyDefinition('A')}
        )


def test_malformed_dependencies():
    with pytest.raises(
        DagsterInvalidDefinitionError,
        match='Expected IDependencyDefinition for solid "B" input "b_input"',
    ):
        PipelineDefinition(
            solid_defs=solid_a_b_list(),
            dependencies={'B': {'b_input': {'b_input': DependencyDefinition('A')}}},
        )


def test_list_dependencies():
    with pytest.raises(
        DagsterInvalidDefinitionError, match='The expected type for "dependencies" is dict'
    ):
        PipelineDefinition(solid_defs=solid_a_b_list(), dependencies=[])


def test_double_type_name():
    @solid(config_field=Field(NamedDict('SomeTypeName', {'some_field': Field(String)})))
    def solid_one(_context):
        raise Exception('should not execute')

    @solid(config_field=Field(NamedDict('SomeTypeName', {'another_field': Field(String)})))
    def solid_two(_context):
        raise Exception('should not execute')

    with pytest.raises(DagsterInvalidDefinitionError) as exc_info:
        create_environment_schema(PipelineDefinition(solid_defs=[solid_one, solid_two]))

    assert str(exc_info.value) == (
        'Type names must be unique. You have constructed two different instances of '
        'types with the same name "SomeTypeName".'
    )


def test_double_type_key():
    class NameOneType(_ConfigHasFields):
        def __init__(self):
            super(NameOneType, self).__init__(
                key='Key', name='NameOne', kind=ConfigTypeKind.DICT, fields={'foo': Field(Int)}
            )

    class NameOneTypePrime(_ConfigHasFields):
        def __init__(self):
            super(NameOneTypePrime, self).__init__(
                key='Key', name='NameOnePrime', kind=ConfigTypeKind.DICT, fields={'bar': Field(Int)}
            )

    @solid(config_field=Field(NameOneType))
    def solid_one(_context):
        raise Exception('should not execute')

    @solid(config_field=Field(NameOneTypePrime))
    def solid_two(_context):
        raise Exception('should not execute')

    with pytest.raises(DagsterInvalidDefinitionError) as exc_info:
        create_environment_schema(PipelineDefinition(solid_defs=[solid_one, solid_two]))

    assert str(exc_info.value) == (
        'Type keys must be unique. You have constructed two different instances of types '
        'with the same key "Key".'
    )


def test_pass_config_type_to_field_error_solid_definition():
    with pytest.raises(DagsterInvalidDefinitionError) as exc_info:

        @solid(config_field=Dict({'val': Field(Int)}))
        def a_solid(_context):
            pass

        assert a_solid  # fool lint

    assert str(exc_info.value) == (
        'You have passed a config type "{ val: Int }" in the parameter "config_field" '
        'of a SolidDefinition or @solid named "a_solid". You have '
        'likely forgot to wrap this type in a Field.'
    )


def test_pass_unrelated_type_to_field_error_solid_definition():

    with pytest.raises(DagsterInvalidDefinitionError) as exc_info:

        @solid(config_field='nope')
        def a_solid(_context):
            pass

        assert a_solid  # fool lint

    assert str(exc_info.value) == (
        'You have passed an object \'nope\' of incorrect type "str" in the parameter '
        '"config_field" of a SolidDefinition or @solid named "a_solid" where a Field '
        'was expected.'
    )


def test_pass_config_type_to_field_error_resource_definition():
    with pytest.raises(DagsterInvalidDefinitionError) as exc_info:
        ResourceDefinition(resource_fn=lambda: None, config_field=Dict({'val': Field(Int)}))

    assert str(exc_info.value) == (
        'You have passed a config type "{ val: Int }" in the parameter "config_field" of a '
        'ResourceDefinition or @resource. You have likely forgot to '
        'wrap this type in a Field.'
    )


def test_pass_unrelated_type_to_field_error_resource_definition():
    with pytest.raises(DagsterInvalidDefinitionError) as exc_info:
        ResourceDefinition(resource_fn=lambda: None, config_field='wut')

    assert str(exc_info.value) == (
        'You have passed an object \'wut\' of incorrect type "str" in the parameter '
        '"config_field" of a ResourceDefinition or @resource where a Field was expected.'
    )


def test_pass_incorrect_thing_to_field():
    with pytest.raises(DagsterInvalidDefinitionError) as exc_info:
        Field('nope')

    assert str(exc_info.value) == (
        'Attempted to pass \'nope\' to a Field that expects a valid dagster type '
        'usable in config (e.g. Dict, Int, String et al).'
    )


def test_invalid_dict_field():
    with pytest.raises(DagsterInvalidDefinitionError) as exc_info:
        Dict({'val': Int, 'another_val': Field(Int)})

    assert str(exc_info.value) == (
        'You have passed a config type "Int" in the parameter "fields" and it is '
        'in the "val" entry of that dict. It is from a Dict with fields '
        '[\'another_val\', \'val\']. You have likely '
        'forgot to wrap this type in a Field.'
    )


def test_invalid_named_dict_field():
    with pytest.raises(DagsterInvalidDefinitionError) as exc_info:
        NamedDict('some_named_dict', {'val': Int, 'another_val': Field(Int)})

    assert str(exc_info.value) == (
        'You have passed a config type "Int" in the parameter "fields" and it is '
        'in the "val" entry of that dict. It is from a NamedDict named '
        '"some_named_dict" with fields [\'another_val\', \'val\']. You '
        'have likely forgot to wrap this type in a Field.'
    )


def test_invalid_permissive_dict_field():

    with pytest.raises(DagsterInvalidDefinitionError) as exc_info:
        PermissiveDict({'val': Int, 'another_val': Field(Int)})

    assert str(exc_info.value) == (
        'You have passed a config type "Int" in the parameter "fields" and it is '
        'in the "val" entry of that dict. It is from a PermissiveDict with fields '
        '[\'another_val\', \'val\']. You have likely '
        'forgot to wrap this type in a Field.'
    )


def test_invalid_selector_field():
    with pytest.raises(DagsterInvalidDefinitionError) as exc_info:
        Selector({'val': Int})

    assert str(exc_info.value) == (
        'You have passed a config type "Int" in the parameter "fields" and it is '
        'in the "val" entry of that dict. It is from a Selector with fields '
        '[\'val\']. You have likely forgot to wrap this type in a Field.'
    )


def test_invalid_named_selector_field():
    with pytest.raises(DagsterInvalidDefinitionError) as exc_info:
        NamedSelector('some_selector', {'val': Int})

    assert str(exc_info.value) == (
        'You have passed a config type "Int" in the parameter "fields" and it is '
        'in the "val" entry of that dict. It is from a NamedSelector named '
        '"some_selector" with fields [\'val\']. You '
        'have likely forgot to wrap this type in a Field.'
    )


def test_bad_output_definition():
    with pytest.raises(
        DagsterInvalidDefinitionError,
        match=re.escape(
            'Invalid type: dagster_type must be a Python type, a type constructed using '
            'python.typing, a type imported from the dagster module, or a class annotated using '
            'as_dagster_type or @dagster_type: got foo'
        ),
    ):
        _output = OutputDefinition('foo')

    # Test the case where the object is not hashable
    with pytest.raises(
        DagsterInvalidDefinitionError,
        match=re.escape(
            'Invalid type: dagster_type must be a Python type, a type constructed using '
            'python.typing, a type imported from the dagster module, or a class annotated using '
            'as_dagster_type or @dagster_type: got {\'foo\': \'bar\'}, which isn\'t hashable. '
            'Did you pass an instance of a type instead of the type?'
        ),
    ):
        _output = OutputDefinition({'foo': 'bar'})

    # Test the case where the object throws in __nonzero__, e.g. pandas.DataFrame
    class Exotic(object):
        def __nonzero__(self):
            raise ValueError('Love too break the core Python APIs in widely-used libraries')

    with pytest.raises(
        DagsterInvalidDefinitionError,
        match=re.escape(
            'Invalid type: dagster_type must be a Python type, a type constructed using '
            'python.typing, a type imported from the dagster module, or a class annotated using '
            'as_dagster_type or @dagster_type: got '
            '<dagster_tests.core_tests.definitions_tests.test_definition_errors'
        )
        + '('  # py27
        + re.escape('.test_bad_output_definition.<locals>')
        + ')?'
        + re.escape('.Exotic object'),
    ):
        _output = OutputDefinition(Exotic())



from collections import namedtuple

from dagster import check
from dagster.core.definitions.environment_configs import is_solid_container_config
from dagster.core.definitions.pipeline import PipelineDefinition
from dagster.core.execution.config import IRunConfig
from dagster.core.types.field import Field

from ..config import ConfigType


class TraversalContext(
    namedtuple(
        '_TraversalContext', 'config_type config_value stack pipeline run_config seen_handles'
    )
):
    def __new__(
        cls, config_type, config_value, stack, pipeline, run_config=None, seen_handles=None
    ):
        from .stack import EvaluationStack

        return super(TraversalContext, cls).__new__(
            cls,
            check.inst_param(config_type, 'config_type', ConfigType),
            config_value,  # arbitrary value, can't check
            check.inst_param(stack, 'stack', EvaluationStack),
            check.opt_inst_param(pipeline, 'pipeline', PipelineDefinition),
            check.opt_inst_param(
                run_config, 'run_config', IRunConfig
            ),  # Can be either RunConfig or PipelineRun
            check.opt_list_param(seen_handles, 'seen_handles'),
        )

    def for_list(self, index, item):
        check.int_param(index, 'index')
        return self._replace(
            config_type=self.config_type.inner_type,
            config_value=item,
            stack=self.stack.for_list_index(index),
        )

    def for_set(self, index, item):
        check.int_param(index, 'index')
        return self._replace(
            config_type=self.config_type.inner_type,
            config_value=item,
            stack=self.stack.for_set_index(index),
        )

    def for_tuple(self, index, item):
        return self._replace(
            config_type=self.config_type.tuple_types[index],
            config_value=item,
            stack=self.stack.for_tuple_index(index),
        )

    def for_field(self, field_def, key, value):
        check.inst_param(field_def, 'field_def', Field)
        check.str_param(key, 'key')
        return self._replace(
            config_type=field_def.config_type,
            config_value=value,
            stack=self.stack.for_field(key, field_def),
        )

    def for_nullable_inner_type(self):
        return self._replace(config_type=self.config_type.inner_type)

    def for_mapped_composite_config(self, handle, mapped_config_value):
        return self._replace(
            config_type=self.config_type.child_solids_config_field.config_type,
            config_value=mapped_config_value,
            seen_handles=self.seen_handles + [handle],
        )

    def new_context_with_handle(self, handle):
        check.invariant(
            is_solid_container_config(self.config_type),
            'Not a solid container config schema; this is intended to create a new context for a '
            'solid container',
        )
        return self._replace(seen_handles=self.seen_handles + [handle])



from dagster.utils import file_relative_path, safe_isfile, script_relative_path


def test_safe_isfile():
    assert safe_isfile(script_relative_path('test_file_utils.py'))
    assert not safe_isfile(script_relative_path('not_a_file.py'))


def test_script_relative_path_file_relative_path_equiv():
    assert script_relative_path('foo') == file_relative_path(__file__, 'foo')



import sys
from collections import namedtuple
from contextlib import contextmanager

from contextlib2 import ExitStack

from dagster import check
from dagster.core.definitions.handle import ExecutionTargetHandle
from dagster.core.definitions.pipeline import PipelineDefinition
from dagster.core.definitions.resource import ScopedResourcesBuilder
from dagster.core.definitions.system_storage import SystemStorageData
from dagster.core.engine.init import InitExecutorContext
from dagster.core.errors import (
    DagsterError,
    DagsterResourceFunctionError,
    DagsterUserCodeExecutionError,
    user_code_error_boundary,
)
from dagster.core.events import DagsterEvent, PipelineInitFailureData
from dagster.core.execution.config import ExecutorConfig
from dagster.core.instance import DagsterInstance
from dagster.core.log_manager import DagsterLogManager
from dagster.core.storage.init import InitSystemStorageContext
from dagster.core.storage.pipeline_run import PipelineRun
from dagster.core.storage.type_storage import construct_type_storage_plugin_registry
from dagster.core.system_config.objects import EnvironmentConfig
from dagster.loggers import default_loggers, default_system_loggers
from dagster.utils import ensure_gen, merge_dicts
from dagster.utils.error import serializable_error_info_from_exc_info

from .context.init import InitResourceContext
from .context.logger import InitLoggerContext
from .context.system import SystemPipelineExecutionContext, SystemPipelineExecutionContextData


@contextmanager
def create_resource_builder(pipeline_def, environment_config, pipeline_run, log_manager):
    check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition)
    check.inst_param(environment_config, 'environment_config', EnvironmentConfig)
    check.inst_param(pipeline_run, 'pipeline_run', PipelineRun)
    check.inst_param(log_manager, 'log_manager', DagsterLogManager)

    resources_stack = ResourcesStack(pipeline_def, environment_config, pipeline_run, log_manager)
    yield resources_stack.create()
    resources_stack.teardown()


def construct_system_storage_data(storage_init_context):
    return storage_init_context.system_storage_def.system_storage_creation_fn(storage_init_context)


def system_storage_def_from_config(mode_definition, environment_config):
    for system_storage_def in mode_definition.system_storage_defs:
        if system_storage_def.name == environment_config.storage.system_storage_name:
            return system_storage_def

    check.failed(
        'Could not find storage mode {}. Should have be caught by config system'.format(
            environment_config.storage.system_storage_name
        )
    )


def construct_executor_config(executor_init_context):
    return executor_init_context.executor_def.executor_creation_fn(executor_init_context)


def executor_def_from_config(mode_definition, environment_config):
    for executor_def in mode_definition.executor_defs:
        if executor_def.name == environment_config.execution.execution_engine_name:
            return executor_def

    check.failed(
        'Could not find executor {}. Should have be caught by config system'.format(
            environment_config.execution.execution_engine_name
        )
    )


# This represents all the data that is passed *into* context creation process.
# The remainder of the objects generated (e.g. loggers, resources) are created
# using user-defined code that may fail at runtime and result in the emission
# of a pipeline init failure events. The data in this object are passed all
# over the place during the context creation process so grouping here for
# ease of argument passing etc.
ContextCreationData = namedtuple(
    'ContextCreationData',
    'pipeline_def environment_config pipeline_run mode_def system_storage_def '
    'execution_target_handle executor_def instance',
)


def create_context_creation_data(pipeline_def, environment_dict, pipeline_run, instance):
    environment_config = EnvironmentConfig.build(pipeline_def, environment_dict, pipeline_run)

    mode_def = pipeline_def.get_mode_definition(pipeline_run.mode)
    system_storage_def = system_storage_def_from_config(mode_def, environment_config)
    executor_def = executor_def_from_config(mode_def, environment_config)

    execution_target_handle, _ = ExecutionTargetHandle.get_handle(pipeline_def)
    return ContextCreationData(
        pipeline_def=pipeline_def,
        environment_config=environment_config,
        pipeline_run=pipeline_run,
        mode_def=mode_def,
        system_storage_def=system_storage_def,
        execution_target_handle=execution_target_handle,
        executor_def=executor_def,
        instance=instance,
    )


@contextmanager
def scoped_pipeline_context(
    pipeline_def,
    environment_dict,
    pipeline_run,
    instance,
    system_storage_data=None,
    scoped_resources_builder_cm=create_resource_builder,
    raise_on_error=False,
):
    check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition)
    check.dict_param(environment_dict, 'environment_dict', key_type=str)
    check.inst_param(pipeline_run, 'pipeline_run', PipelineRun)
    check.inst_param(instance, 'instance', DagsterInstance)
    check.opt_inst_param(system_storage_data, 'system_storage_data', SystemStorageData)

    context_creation_data = create_context_creation_data(
        pipeline_def, environment_dict, pipeline_run, instance
    )

    executor_config = create_executor_config(context_creation_data)

    # After this try block, a Dagster exception thrown will result in a pipeline init failure event.
    pipeline_context = None
    try:
        executor_config.check_requirements(instance, context_creation_data.system_storage_def)

        log_manager = create_log_manager(context_creation_data)

        with scoped_resources_builder_cm(
            context_creation_data.pipeline_def,
            context_creation_data.environment_config,
            context_creation_data.pipeline_run,
            log_manager,
        ) as scoped_resources_builder:

            system_storage_data = create_system_storage_data(
                context_creation_data, system_storage_data, scoped_resources_builder
            )

            pipeline_context = construct_pipeline_execution_context(
                context_creation_data=context_creation_data,
                scoped_resources_builder=scoped_resources_builder,
                system_storage_data=system_storage_data,
                log_manager=log_manager,
                executor_config=executor_config,
                raise_on_error=raise_on_error,
            )
            yield pipeline_context

    except DagsterError as dagster_error:
        # only yield an init failure event if we haven't already yielded context
        if pipeline_context is None:
            user_facing_exc_info = (
                # pylint does not know original_exc_info exists is is_user_code_error is true
                # pylint: disable=no-member
                dagster_error.original_exc_info
                if dagster_error.is_user_code_error
                else sys.exc_info()
            )

            error_info = serializable_error_info_from_exc_info(user_facing_exc_info)
            yield DagsterEvent.pipeline_init_failure(
                pipeline_name=pipeline_def.name,
                failure_data=PipelineInitFailureData(error=error_info),
                log_manager=_create_context_free_log_manager(instance, pipeline_run, pipeline_def),
            )

            if raise_on_error:
                raise dagster_error

        # if we've caught an error after context init we're in a problematic state and should just raise
        else:
            raise dagster_error


def create_system_storage_data(
    context_creation_data, system_storage_data, scoped_resources_builder
):
    check.inst_param(context_creation_data, 'context_creation_data', ContextCreationData)

    environment_config, pipeline_def, system_storage_def, pipeline_run = (
        context_creation_data.environment_config,
        context_creation_data.pipeline_def,
        context_creation_data.system_storage_def,
        context_creation_data.pipeline_run,
    )

    system_storage_data = (
        system_storage_data
        if system_storage_data
        else construct_system_storage_data(
            InitSystemStorageContext(
                pipeline_def=pipeline_def,
                mode_def=context_creation_data.mode_def,
                system_storage_def=system_storage_def,
                system_storage_config=environment_config.storage.system_storage_config,
                pipeline_run=pipeline_run,
                instance=context_creation_data.instance,
                environment_config=environment_config,
                type_storage_plugin_registry=construct_type_storage_plugin_registry(
                    pipeline_def, system_storage_def
                ),
                resources=scoped_resources_builder.build(
                    # currently provide default for resource mapping
                    lambda resources, resources_deps: {r: resources.get(r) for r in resources_deps},
                    context_creation_data.system_storage_def.required_resource_keys,
                ),
            )
        )
    )
    return system_storage_data


def create_executor_config(context_creation_data):
    check.inst_param(context_creation_data, 'context_creation_data', ContextCreationData)

    environment_config, pipeline_def, executor_def, pipeline_run = (
        context_creation_data.environment_config,
        context_creation_data.pipeline_def,
        context_creation_data.executor_def,
        context_creation_data.pipeline_run,
    )

    return construct_executor_config(
        InitExecutorContext(
            pipeline_def=pipeline_def,
            mode_def=context_creation_data.mode_def,
            executor_def=executor_def,
            pipeline_run=pipeline_run,
            environment_config=environment_config,
            executor_config=environment_config.execution.execution_engine_config,
        )
    )


def construct_pipeline_execution_context(
    context_creation_data,
    scoped_resources_builder,
    system_storage_data,
    log_manager,
    executor_config,
    raise_on_error,
):
    check.inst_param(context_creation_data, 'context_creation_data', ContextCreationData)
    scoped_resources_builder = check.inst_param(
        scoped_resources_builder if scoped_resources_builder else ScopedResourcesBuilder(),
        'scoped_resources_builder',
        ScopedResourcesBuilder,
    )
    check.inst_param(system_storage_data, 'system_storage_data', SystemStorageData)
    check.inst_param(log_manager, 'log_manager', DagsterLogManager)
    check.inst_param(executor_config, 'executor_config', ExecutorConfig)

    return SystemPipelineExecutionContext(
        SystemPipelineExecutionContextData(
            pipeline_def=context_creation_data.pipeline_def,
            mode_def=context_creation_data.mode_def,
            system_storage_def=context_creation_data.system_storage_def,
            pipeline_run=context_creation_data.pipeline_run,
            scoped_resources_builder=scoped_resources_builder,
            environment_config=context_creation_data.environment_config,
            instance=context_creation_data.instance,
            intermediates_manager=system_storage_data.intermediates_manager,
            file_manager=system_storage_data.file_manager,
            execution_target_handle=context_creation_data.execution_target_handle,
            executor_config=executor_config,
            raise_on_error=raise_on_error,
        ),
        log_manager=log_manager,
    )


class ResourcesStack(object):
    # Wraps an ExitStack to support execution in environments like Dagstermill, where we can't
    # wrap solid execution/the pipeline execution context lifecycle in an ordinary Python context
    # manager (because notebook execution is cell-by-cell in the Jupyter kernel, a subprocess we
    # don't directly control). In these environments we need to manually create and teardown
    # resources.
    def __init__(self, pipeline_def, environment_config, pipeline_run, log_manager):
        check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition)
        check.inst_param(environment_config, 'environment_config', EnvironmentConfig)
        check.inst_param(pipeline_run, 'pipeline_run', PipelineRun)
        check.inst_param(log_manager, 'log_manager', DagsterLogManager)

        self.resource_instances = {}
        self.mode_definition = pipeline_def.get_mode_definition(pipeline_run.mode)
        self.pipeline_def = pipeline_def
        self.environment_config = environment_config
        self.pipeline_run = pipeline_run
        self.log_manager = log_manager
        self.stack = ExitStack()

    def create(self):
        for resource_name, resource_def in sorted(self.mode_definition.resource_defs.items()):
            user_fn = create_resource_fn_lambda(
                self.pipeline_def,
                resource_def,
                self.environment_config.resources.get(resource_name, {}).get('config'),
                self.pipeline_run.run_id,
                self.log_manager,
            )

            def _create_msg_fn(rn):
                return lambda: 'Error executing resource_fn on ResourceDefinition {name}'.format(
                    name=rn
                )

            resource_obj = self.stack.enter_context(
                user_code_context_manager(
                    user_fn, DagsterResourceFunctionError, _create_msg_fn(resource_name)
                )
            )

            self.resource_instances[resource_name] = resource_obj
        return ScopedResourcesBuilder(self.resource_instances)

    def teardown(self):
        self.stack.close()


def create_resource_fn_lambda(pipeline_def, resource_def, resource_config, run_id, log_manager):
    return lambda: resource_def.resource_fn(
        InitResourceContext(
            pipeline_def=pipeline_def,
            resource_def=resource_def,
            resource_config=resource_config,
            run_id=run_id,
            log_manager=log_manager,
        )
    )


def create_log_manager(context_creation_data):
    check.inst_param(context_creation_data, 'context_creation_data', ContextCreationData)

    pipeline_def, mode_def, environment_config, pipeline_run = (
        context_creation_data.pipeline_def,
        context_creation_data.mode_def,
        context_creation_data.environment_config,
        context_creation_data.pipeline_run,
    )

    loggers = []
    for logger_key, logger_def in mode_def.loggers.items() or default_loggers().items():
        if logger_key in environment_config.loggers:
            loggers.append(
                logger_def.logger_fn(
                    InitLoggerContext(
                        environment_config.loggers.get(logger_key, {}).get('config'),
                        pipeline_def,
                        logger_def,
                        pipeline_run.run_id,
                    )
                )
            )

    if not loggers:
        for (logger_def, logger_config) in default_system_loggers():
            loggers.append(
                logger_def.logger_fn(
                    InitLoggerContext(logger_config, pipeline_def, logger_def, pipeline_run.run_id)
                )
            )

    # should this be first in loggers list?
    loggers.append(context_creation_data.instance.get_logger())

    return DagsterLogManager(
        run_id=pipeline_run.run_id,
        logging_tags=get_logging_tags(pipeline_run, context_creation_data.pipeline_def),
        loggers=loggers,
    )


def _create_context_free_log_manager(instance, pipeline_run, pipeline_def):
    '''In the event of pipeline initialization failure, we want to be able to log the failure
    without a dependency on the ExecutionContext to initialize DagsterLogManager.
    Args:
        pipeline_run (dagster.core.storage.pipeline_run.PipelineRun)
        pipeline_def (dagster.definitions.PipelineDefinition)
    '''
    check.inst_param(instance, 'instance', DagsterInstance)
    check.inst_param(pipeline_run, 'pipeline_run', PipelineRun)
    check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition)

    loggers = [instance.get_logger()]
    # Use the default logger
    for (logger_def, logger_config) in default_system_loggers():
        loggers += [
            logger_def.logger_fn(
                InitLoggerContext(logger_config, pipeline_def, logger_def, pipeline_run.run_id)
            )
        ]

    return DagsterLogManager(
        pipeline_run.run_id, get_logging_tags(pipeline_run, pipeline_def), loggers
    )


@contextmanager
def user_code_context_manager(user_fn, error_cls, msg_fn):
    '''Wraps the output of a user provided function that may yield or return a value and
    returns a generator that asserts it only yields a single value.
    '''
    check.callable_param(user_fn, 'user_fn')
    check.subclass_param(error_cls, 'error_cls', DagsterUserCodeExecutionError)

    with user_code_error_boundary(error_cls, msg_fn):
        thing_or_gen = user_fn()
        gen = ensure_gen(thing_or_gen)

        try:
            thing = next(gen)
        except StopIteration:
            check.failed('Must yield one item. You did not yield anything.')

        yield thing

        stopped = False

        try:
            next(gen)
        except StopIteration:
            stopped = True

        check.invariant(stopped, 'Must yield one item. Yielded more than one item')


def get_logging_tags(pipeline_run, pipeline):
    check.opt_inst_param(pipeline_run, 'pipeline_run', PipelineRun)
    check.inst_param(pipeline, 'pipeline', PipelineDefinition)

    return merge_dicts(
        {'pipeline': pipeline.name}, pipeline_run.tags if pipeline_run and pipeline_run.tags else {}
    )



'''
Serialization & deserialization for Dagster objects.

Why have custom serialization?

* Default json serialization doesn't work well on namedtuples, which we use extensively to create
  immutable value types. Namedtuples serialize like tuples as flat lists.
* Explicit whitelisting should help ensure we are only persisting or communicating across a
  serialization boundary the types we expect to.

Why not pickle?

* This isn't meant to replace pickle in the conditions that pickle is reasonable to use
  (in memory, not human readble, etc) just handle the json case effectively.
'''
import importlib
from abc import ABCMeta, abstractmethod, abstractproperty
from collections import namedtuple
from enum import Enum

import six
import yaml

from dagster import check, seven

_WHITELISTED_TUPLE_MAP = {}
_WHITELISTED_ENUM_MAP = {}


def _whitelist_for_serdes(enum_map, tuple_map):
    def __whitelist_for_serdes(klass):
        if issubclass(klass, Enum):
            enum_map[klass.__name__] = klass
        elif issubclass(klass, tuple):
            tuple_map[klass.__name__] = klass
        else:
            check.failed('Can not whitelist class {klass} for serdes'.format(klass=klass))
        return klass

    return __whitelist_for_serdes


def whitelist_for_serdes(klass):
    return _whitelist_for_serdes(enum_map=_WHITELISTED_ENUM_MAP, tuple_map=_WHITELISTED_TUPLE_MAP)(
        klass
    )


def pack_value(val):
    return _pack_value(val, enum_map=_WHITELISTED_ENUM_MAP, tuple_map=_WHITELISTED_TUPLE_MAP)


def _pack_value(val, enum_map, tuple_map):
    if isinstance(val, list):
        return [_pack_value(i, enum_map, tuple_map) for i in val]
    if isinstance(val, tuple):
        klass_name = val.__class__.__name__
        check.invariant(
            klass_name in tuple_map,
            'Can only serialize whitelisted namedtuples, recieved {}'.format(klass_name),
        )
        base_dict = {
            key: _pack_value(value, enum_map, tuple_map) for key, value in val._asdict().items()
        }
        base_dict['__class__'] = klass_name
        return base_dict
    if isinstance(val, Enum):
        klass_name = val.__class__.__name__
        check.invariant(
            klass_name in enum_map,
            'Can only serialize whitelisted Enums, recieved {}'.format(klass_name),
        )
        return {'__enum__': str(val)}
    if isinstance(val, dict):
        return {key: _pack_value(value, enum_map, tuple_map) for key, value in val.items()}

    return val


def _serialize_dagster_namedtuple(nt, enum_map, tuple_map):
    return seven.json.dumps(_pack_value(nt, enum_map, tuple_map))


def serialize_dagster_namedtuple(nt):
    return _serialize_dagster_namedtuple(
        nt, enum_map=_WHITELISTED_ENUM_MAP, tuple_map=_WHITELISTED_TUPLE_MAP
    )


def unpack_value(val):
    return _unpack_value(val, enum_map=_WHITELISTED_ENUM_MAP, tuple_map=_WHITELISTED_TUPLE_MAP)


def _unpack_value(val, enum_map, tuple_map):
    if isinstance(val, list):
        return [_unpack_value(i, enum_map, tuple_map) for i in val]
    if isinstance(val, dict) and val.get('__class__'):
        klass_name = val.pop('__class__')
        klass = tuple_map[klass_name]
        unpacked_val = {
            key: _unpack_value(value, enum_map, tuple_map) for key, value in val.items()
        }

        # Naively implements backwards compatibility by filtering arguments that aren't present in
        # the constructor. If a property is present in the serialized object, but doesn't exist in
        # the version of the class loaded into memory, that property will be completely ignored.
        args_for_class = seven.get_args(klass)
        filtered_val = (
            {k: v for k, v in unpacked_val.items() if k in args_for_class}
            if args_for_class
            else unpacked_val
        )
        return klass(**filtered_val)
    if isinstance(val, dict) and val.get('__enum__'):
        name, member = val['__enum__'].split('.')
        return getattr(enum_map[name], member)
    if isinstance(val, dict):
        return {key: _unpack_value(value, enum_map, tuple_map) for key, value in val.items()}

    return val


def deserialize_json_to_dagster_namedtuple(json_str):
    return _deserialize_json_to_dagster_namedtuple(
        check.str_param(json_str, 'json_str'),
        enum_map=_WHITELISTED_ENUM_MAP,
        tuple_map=_WHITELISTED_TUPLE_MAP,
    )


def _deserialize_json_to_dagster_namedtuple(json_str, enum_map, tuple_map):
    return _unpack_value(seven.json.loads(json_str), enum_map=enum_map, tuple_map=tuple_map)


@whitelist_for_serdes
class ConfigurableClassData(
    namedtuple('_ConfigurableClassData', 'module_name class_name config_yaml')
):
    '''Serializable tuple describing where to find a class and the config fragment that should
    be used to instantiate it.

    Classes serialized in this way should implement dagster.serdes.ConfigurableClass.
    '''

    def __new__(cls, module_name, class_name, config_yaml):
        return super(ConfigurableClassData, cls).__new__(
            cls,
            check.str_param(module_name, 'module_name'),
            check.str_param(class_name, 'class_name'),
            check.str_param(config_yaml, 'config_yaml'),
        )

    def info_str(self, prefix=''):
        return (
            '{p}module: {module}\n'
            '{p}class: {cls}\n'
            '{p}config:\n'
            '{p}  {config}'.format(
                p=prefix, module=self.module_name, cls=self.class_name, config=self.config_yaml
            )
        )

    def rehydrate(self, **constructor_kwargs):
        from dagster.core.errors import DagsterInvalidConfigError
        from dagster.core.types.evaluator import evaluate_config

        try:
            module = importlib.import_module(self.module_name)
        except seven.ModuleNotFoundError:
            check.failed(
                'Couldn\'t import module {module_name} when attempting to rehydrate the '
                'configurable class {configurable_class}'.format(
                    module_name=self.module_name,
                    configurable_class=self.module_name + '.' + self.class_name,
                )
            )
        try:
            klass = getattr(module, self.class_name)
        except AttributeError:
            check.failed(
                'Couldn\'t find class {class_name} in module when attempting to rehydrate the '
                'configurable class {configurable_class}'.format(
                    class_name=self.class_name,
                    configurable_class=self.module_name + '.' + self.class_name,
                )
            )

        if not issubclass(klass, ConfigurableClass):
            raise check.CheckError(
                klass,
                'class {class_name} in module {module_name}'.format(
                    class_name=self.class_name, module_name=self.module_name
                ),
                ConfigurableClass,
            )

        config_dict = yaml.load(self.config_yaml)
        result = evaluate_config(klass.config_type().inst(), config_dict)
        if not result.success:
            raise DagsterInvalidConfigError(None, result.errors, config_dict)
        return klass.from_config_value(self, result.value, **constructor_kwargs)


class ConfigurableClass(six.with_metaclass(ABCMeta)):
    '''Abstract mixin for classes that can be rehydrated from config.

    This supports a powerful plugin pattern which avoids both a) a lengthy, hard-to-synchronize list
    of conditional imports / optional extras_requires in dagster core and b) a magic directory or
    file in which third parties can place plugin packages. Instead, the intention is to make, e.g.,
    run storage, pluggable with a config chunk like:

        run_storage:
            module: very_cool_package.run_storage
            class: SplendidRunStorage
            config:
                magic_word: "quux"
    
    This same pattern should eventually be viable for other system components, e.g. engines.

    The ConfigurableClass mixin provides the necessary hooks for classes to be instantiated from
    an instance of ConfigurableClassData.

    Pieces of the Dagster system which we wish to make pluggable in this way should consume a config
    type such as:

        SystemNamedDict(
            name,
            {'module': Field(String), 'class': Field(String), 'config': Field(PermissiveDict())},
        )
    '''

    @abstractproperty
    def inst_data(self):
        '''
        Subclass must be able to return the inst_data as a property if it has been constructed
        through the from_config_value code path.
        '''

    @classmethod
    @abstractmethod
    def config_type(cls):
        '''dagster.ConfigType: The config type against which to validate a config yaml fragment
        serialized in an instance of ConfigurableClassData.
        
        This is usually an instance of dagster.core.definitions.environment_configs.SystemNamedDict.
        '''

    @staticmethod
    @abstractmethod
    def from_config_value(inst_data, config_value, **kwargs):
        '''New up an instance of the ConfigurableClass from a validated config value.

        Called by ConfigurableClassData.rehydrate.
        
        Args:
            config_value (dict): The validated config value to use. Typically this should be the
                `value` attribute of a dagster.core.types.evaluator.evaluation.EvaluateValueResult.


        A common pattern is for the implementation to splat the config_value with the kwargs, to
        align with the signature of the ConfigurableClass's constructor and allow overrides:

            @staticmethod
            def from_config_value(inst_data, config_value, **kwargs):
                return MyConfigurableClass(inst_data=inst_data, **dict(config_value, **kwargs))

        '''



'''
This module contains serializable classes that contain all the meta information
in our definitions and type systems. The purpose to be able to power our graphql
interface without having access to all the definitions in and types in process.

This will have a number of uses, but the most immediately germane are:

1) Persist *historical* pipeline and repository structures. This
will enable, in the short term, for the user to be able to go to a historical
run and view the meta information at that point in time.
2) Cache the meta information about a repository when, for example, interacting
with a repository that is not in the same process but is instead resident in
a container. This way one does not have query the meta information about
a pipeline/repository in the critical path of a user interaction.

I suspect also that the ability to hash the entire construct of meta information
to identify the repository will also end up being quite useful.

    -- schrockn 11-22-2019

'''



import typing

from dagster import check
from dagster.core.errors import DagsterInvalidDefinitionError

from .builtin_enum import BuiltinEnum
from .typing_api import (
    get_optional_inner_type,
    is_closed_python_optional_type,
    is_python_list_type,
    is_python_set_type,
    is_python_tuple_type,
)


class WrappingType(object):
    def __init__(self, inner_type):
        # Cannot check inner_type because of circular references and no fwd declarations

        if inner_type == BuiltinEnum.NOTHING:
            raise DagsterInvalidDefinitionError(
                'Type Nothing can not be wrapped in List or Optional'
            )

        self.inner_type = inner_type


class WrappingListType(WrappingType):
    pass


class WrappingSetType(WrappingType):
    pass


class WrappingTupleType(WrappingType):
    pass


class WrappingNullableType(WrappingType):
    pass


def remap_to_dagster_list_type(ttype):
    check.invariant(is_python_list_type(ttype), 'type must pass is_python_list_type check')
    if ttype == list or ttype == typing.List:
        return WrappingListType(BuiltinEnum.ANY)
    return WrappingListType(ttype.__args__[0])


def remap_to_dagster_set_type(ttype):
    check.invariant(is_python_set_type(ttype), 'type must pass is_python_set_type check')
    if ttype == set or ttype == typing.Set:
        return WrappingSetType(BuiltinEnum.ANY)
    return WrappingSetType(ttype.__args__[0])


def remap_to_dagster_tuple_type(ttype):
    check.invariant(is_python_tuple_type(ttype), 'type must pass is_python_tuple_type check')
    if ttype == tuple or ttype == typing.Tuple:
        return WrappingTupleType(None)
    return WrappingTupleType(ttype.__args__)


def remap_to_dagster_optional_type(ttype):
    check.invariant(
        is_closed_python_optional_type(ttype), 'type must pass is_closed_python_optional_type check'
    )
    return WrappingNullableType(get_optional_inner_type(ttype))


List = typing.List

Optional = typing.Optional
